/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2016-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl;

import java.io.IOException;
import java.net.SocketTimeoutException;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;

import com.is2t.support.net.ssl.natives.SSLSocketSupportNatives;

import ej.error.Message;
import ej.sni.SNI;

public class SSLSocketSupport {

	private static boolean initialized = false;

	static {
		initialize();
	}

	public static void initialize() {
		if (!initialized) {
			int result = SSLSocketSupportNatives.initialize();
			if (result < 0) {
				throw new RuntimeException("SSL initialization failed");
			}
		}
		initialized = true;
	}

	public static int create(int contextid, int fd, String host, boolean autoclose, boolean isClientMode, boolean needClientAuth)
			throws SSLException {

		// call the native with an empty byte array in order to be backward compatible
		// (since the native may not expect a NULL array)
		byte[] hostname = new byte[host == null ? 0 : host.length() + 1];
		int hostLength = hostname.length;
		if (host != null) {
			SNI.toCString(host, hostname);
			hostLength = host.length() + 1; //+1 for null-terminated
		}
		int ret = SSLSocketSupportNatives.create(contextid, fd, hostname, hostLength, autoclose, isClientMode, needClientAuth, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.create(contextid, fd, hostname, hostLength, autoclose, isClientMode, needClientAuth, true);
		}
		if (ret == SSLErrorConstants.J_CREATE_SSL_ERROR) {
			throw new SSLException("Could not create the underlying SSL Socket");
		}

		return ret;
	}

	public static void close(SSLSocketImpl socket, boolean autoclose) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.close(sslid, fd, autoclose, false);
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.close(sslid, fd, autoclose, true);
		}

		if (ret < 0) {
			throw new SSLException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void freeSSL(SSLSocketImpl socket) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		ret = SSLSocketSupportNatives.freeSSL(sslid, false);
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.freeSSL(sslid, true);
		}

		if (ret < 0) {
			throw new SSLException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void initialClientHandShake(SSLSocketImpl socket) throws SSLException {

		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.initialClientHandShake(sslid, fd, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

			ret = SSLSocketSupportNatives.initialClientHandShake(sslid, fd, true);
		}

		if (ret < 0) {
			throw new SSLHandshakeException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void initialServerHandShake(SSLSocketImpl socket) throws SSLException {
		int ret = -1;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.initialServerHandShake(sslid, fd, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

			ret = SSLSocketSupportNatives.initialServerHandShake(sslid, fd, true);
		}

		if (ret < 0) {
			throw new SSLHandshakeException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static int read(SSLSocketImpl socket, byte[] buf, int offset, int length) throws IOException {
		int ret;
		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		ret = SSLSocketSupportNatives.read(sslid, fd, buf, offset, length, false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {

			ret = SSLSocketSupportNatives.read(sslid, fd, buf, offset, length, true);
		}

		if (ret < SSLErrorConstants.J_EOF) {

			if (ret == SSLErrorConstants.J_SOCKET_TIMEOUT) {
				// timeout exception
				throw new SocketTimeoutException();
			}

			// error
			throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
		}

		return ret;
	}

	public static int available(SSLSocketImpl socket) throws IOException {
		int ret = SSLSocketSupportNatives.available(socket.getSSLID(), false);

		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLSocketSupportNatives.available(socket.getSSLID(), true);
		}

		if (ret < 0) {
			throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
		return ret;
	}

	public static int write(SSLSocketImpl socket, byte[] buf, int offset, int length) throws IOException {

		int writtenBytes = 0;

		int sslid = socket.getSSLID();
		int fd = socket.getFD();
		int remaining;
		// while there is data to write, transmit by "IMMORTAL capacity" (at
		// most) chunks
		while ((remaining = (length - writtenBytes)) > 0) { // we did not
			// reach the end
			// of buffer
			// put into IMMORTAL
			int len = Math.min(remaining, buf.length);

			int ret = SSLSocketSupportNatives.write(sslid, fd, buf, offset, len, false);

			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SSLSocketSupportNatives.write(sslid, fd, buf, offset, len, true);
			}
			if (ret <= 0) {
				throw new IOException(Message.at(SSLErrorMessages.getInstance(), ret));
			}
			writtenBytes += ret;
			offset += ret;
		}
		return writtenBytes;
	}

}
